@@ -1,8 +1,9 @@
 use core::char;
 use core::cmp;
+use core::ops::Range;
 use core::str;
 
-use {
+use crate::{
     StreamError,
     StrSpan,
     TextPos,
@@ -59,6 +60,16 @@ impl<'a> From<StrSpan<'a>> for Stream<'a> {
 }
 
 impl<'a> Stream<'a> {
+    /// Creates a new stream from a specified `text` substring.
+    #[inline]
+    pub fn from_substr(text: &'a str, fragment: Range<usize>) -> Self {
+        Stream {
+            pos: fragment.start,
+            end: fragment.end,
+            span: text.into(),
+        }
+    }
+
     /// Returns an underling string span.
     #[inline]
     pub fn span(&self) -> StrSpan<'a> {
@@ -253,6 +264,7 @@ impl<'a> Stream<'a> {
     /// Consumes chars by the predicate and returns them.
     ///
     /// The result can be empty.
+    #[inline]
     pub fn consume_chars<F>(&mut self, f: F) -> Result<StrSpan<'a>>
         where F: Fn(&Stream, char) -> bool
     {
@@ -262,6 +274,7 @@ impl<'a> Stream<'a> {
     }
 
     /// Skips chars by the predicate.
+    #[inline]
     pub fn skip_chars<F>(&mut self, f: F) -> Result<()>
         where F: Fn(&Stream, char) -> bool
     {
@@ -295,62 +308,6 @@ impl<'a> Stream<'a> {
         self.span.slice_region(self.pos, self.end)
     }
 
-    /// Calculates a current absolute position.
-    ///
-    /// This operation is very expensive. Use only for errors.
-    #[inline(never)]
-    pub fn gen_text_pos(&self) -> TextPos {
-        let text = self.span.full_str();
-        let end = self.pos + self.span.start();
-
-        let row = Self::calc_curr_row(text, end);
-        let col = Self::calc_curr_col(text, end);
-        TextPos::new(row, col)
-    }
-
-    /// Calculates an absolute position at `pos`.
-    ///
-    /// This operation is very expensive. Use only for errors.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// let s = xmlparser::Stream::from("text");
-    ///
-    /// assert_eq!(s.gen_text_pos_from(2), xmlparser::TextPos::new(1, 3));
-    /// assert_eq!(s.gen_text_pos_from(9999), xmlparser::TextPos::new(1, 5));
-    /// ```
-    #[inline(never)]
-    pub fn gen_text_pos_from(&self, pos: usize) -> TextPos {
-        let mut s = self.clone();
-        s.pos = cmp::min(pos, s.span.full_str().len());
-        s.gen_text_pos()
-    }
-
-    fn calc_curr_row(text: &str, end: usize) -> u32 {
-        let mut row = 1;
-        for c in &text.as_bytes()[..end] {
-            if *c == b'\n' {
-                row += 1;
-            }
-        }
-
-        row
-    }
-
-    fn calc_curr_col(text: &str, end: usize) -> u32 {
-        let mut col = 1;
-        for c in text[..end].chars().rev() {
-            if c == '\n' {
-                break;
-            } else {
-                col += 1;
-            }
-        }
-
-        col
-    }
-
     /// Skips whitespaces.
     ///
     /// Accepted values: `' ' \n \r \t`.
@@ -421,6 +378,7 @@ impl<'a> Stream<'a> {
         self._consume_reference().map_err(|_| StreamError::InvalidReference)
     }
 
+    #[inline(never)]
     fn _consume_reference(&mut self) -> Result<Reference<'a>> {
         if !self.try_consume_byte(b'&') {
             return Err(StreamError::InvalidReference);
@@ -515,6 +473,7 @@ impl<'a> Stream<'a> {
     /// # Errors
     ///
     /// - `InvalidName` - if name is empty or starts with an invalid char
+    #[inline(never)]
     pub fn consume_qname(&mut self) -> Result<(StrSpan<'a>, StrSpan<'a>)> {
         let start = self.pos();
 
@@ -610,4 +569,60 @@ impl<'a> Stream<'a> {
             Err(StreamError::InvalidQuote(c, self.gen_text_pos()))
         }
     }
+
+    /// Calculates a current absolute position.
+    ///
+    /// This operation is very expensive. Use only for errors.
+    #[inline(never)]
+    pub fn gen_text_pos(&self) -> TextPos {
+        let text = self.span.as_str();
+        let end = self.pos;
+
+        let row = Self::calc_curr_row(text, end);
+        let col = Self::calc_curr_col(text, end);
+        TextPos::new(row, col)
+    }
+
+    /// Calculates an absolute position at `pos`.
+    ///
+    /// This operation is very expensive. Use only for errors.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// let s = xmlparser::Stream::from("text");
+    ///
+    /// assert_eq!(s.gen_text_pos_from(2), xmlparser::TextPos::new(1, 3));
+    /// assert_eq!(s.gen_text_pos_from(9999), xmlparser::TextPos::new(1, 5));
+    /// ```
+    #[inline(never)]
+    pub fn gen_text_pos_from(&self, pos: usize) -> TextPos {
+        let mut s = self.clone();
+        s.pos = cmp::min(pos, s.span.as_str().len());
+        s.gen_text_pos()
+    }
+
+    fn calc_curr_row(text: &str, end: usize) -> u32 {
+        let mut row = 1;
+        for c in &text.as_bytes()[..end] {
+            if *c == b'\n' {
+                row += 1;
+            }
+        }
+
+        row
+    }
+
+    fn calc_curr_col(text: &str, end: usize) -> u32 {
+        let mut col = 1;
+        for c in text[..end].chars().rev() {
+            if c == '\n' {
+                break;
+            } else {
+                col += 1;
+            }
+        }
+
+        col
+    }
 }
